/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include "config_xor.h"

#ifdef HAVE_BROKEN_INCLUDES
#define _ANSI_C_SOURCE
#define _POSIX_SOURCE
#endif


#include <stdio.h>     
#include <stdlib.h>     
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <syslog.h>
#include <errno.h>

/* Required on FreeBSD
 */
#include <sys/types.h>

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
#include <sys/mman.h>
#endif

#if defined(__linux__)
#include <paths.h>
#endif
#ifndef _PATH_CONSOLE
#define _PATH_CONSOLE   "/dev/console"
#endif

#include "samhain.h"
#include "sh_error.h"
#include "sh_utils.h"
#include "sh_unix.h"
#include "sh_tiger.h"
#include "sh_mail.h"
#include "sh_forward.h"

#if defined(WITH_EXTERNAL)
#include "sh_extern.h"
#endif


int  ShDFLevel[SH_ERR_T_END];

struct  errFlagsStruct {
  int           debug;
  int           HaveLog;

  int           loglevel;
  int           loglevel_temp;
  int           printlevel;
  int           maillevel;
  int           exportlevel;
  int           sysloglevel;
  int           externallevel;

  ShErrLevel    severity;
  char          errmsg[SH_MINIBUF+2];
  char          errfile[PATH_MAX+2];
} errFlags;

static int  sh_error_init ();
static int  sh_error_string (char *message,  ShErrLevel severity, 
			     const char *fmt, va_list vl);
static int  sh_error_print (char *message);
static int  sh_error_log (char *message);
static int  sh_error_syslog (ShErrLevel  severity, char *message);


static int  IsInitialized = BAD;

static int  OnlyStderr    = GOOD; /* only log to stderr */


/*********************************************
 *  utility functions for verifying entries
 *********************************************/

int sh_error_verify (char * s)
{
  char * foo;

  if (s[0] == '/')  /* a filename */
    {
      foo = sh_tiger_hash_gpg (s, TIGER_FILE, 0);
      fprintf (stdout, _("%s\n"),  foo);
      SH_FREE(foo);
    }
  else
    {
      fprintf (stdout, _("HASH : |%s|\n"), 
	       sh_tiger_hash (s, TIGER_DATA, sl_strlen(s)));
      fprintf (stdout, _("ORIG : |%s|\n"), s);
    }
  fflush(stdout);
  _exit (EXIT_SUCCESS);
  return 0;
}


int sh_error_logverify (char * s)
{
  SL_TICKET fd;
  int len;
  int count =  0;
  int start = -1;
  char * buf;
  char * bufc;
  char signature[64];
  char key[KEY_LEN+1];
  char timestamp[64];
  char c_cont;

  sh_error_logoff();

  if (s == NULL || sl_strlen(s) >= PATH_MAX)
    {
      fprintf(stderr, _("Bad input %s\n"), s);
      _exit (EXIT_FAILURE);
    }

  /* open the file, then check it 
   */
  if ( SL_ISERROR(fd = sl_open_read (s, SL_NOPRIV)) )
    {
      fprintf(stderr, _("Could not open file %s\n"), s);
      _exit (EXIT_FAILURE);
    }

#ifndef SH_STEALTH
  c_cont = ' ';
#else
  signature[0] = ' ';
  signature[1] = '\0';
  sh_do_encode (signature, 1);
  c_cont = signature[0];
  memset (signature, '\0', 64);
#endif

  buf  = SH_ALLOC( (unsigned long)(2*SH_BUFSIZE+1));
  bufc = SH_ALLOC( (unsigned long)(2*SH_BUFSIZE+1));

  while (1) {

    /* get the log message
     */
    if (sh_unix_getline (fd, buf, (2*SH_BUFSIZE)) < 0) 
      break;


    len = (int) sl_strlen(buf);

#ifdef SH_STEALTH
    if (0 == sl_strncmp (buf, N_("[SOF]"), 5)) 
#else
    if (0 == sl_strncmp (buf, _("[SOF]"),  5)) 
#endif
      {
	/* found start of audit trail, read first line */
	start = 1;
	do {
	  (void) sh_unix_getline (fd, buf, (2*SH_BUFSIZE));
	} while (buf[0] == '\0' || buf[0] == '\n');
	len = (int) sl_strlen(buf);
	++count;
      }
    else if (buf[0] == '\n')
      {
	/* a newline */
	++count;
	continue;
      }
    else if (start == 0)
      {
	/* we are inside an audit trail */
	++count;
      }
    else
      {
	/* no start-of-file found yet */
	continue;
      }

    /* check for continuation line
     */
    while (1 == 1)
      {
	do {
	  (void) sh_unix_getline (fd, bufc, (2*SH_BUFSIZE));
	} while (bufc[0] == '\0' || bufc[0] == '\n');
	++count;
	if (bufc[0] == c_cont) /* a continuation line */
	  {
	    sl_strlcat(buf, "\n", 2*SH_BUFSIZE+1);  /* add the newline */
	    ++len;
	    sl_strlcat(buf, bufc, 2*SH_BUFSIZE+1);
	    len += (int) sl_strlen(bufc);
	  }
	else
	  {
	    sl_strlcpy(signature, bufc, 64);
	    break;
	  }
      }

    /* get starting key from cl 
     */    
    if (start == 1) 
      {

	/* get tiemstamp
	 */
	sl_strlcpy(timestamp, &buf[10], 20);
	timestamp[19] = '\0';
#ifdef SH_STEALTH
	sh_do_decode (timestamp, sl_strlen(timestamp));
#endif

	key[0] = '\0';
	fprintf(stdout, _("New audit trail (%s), enter key: "), timestamp);
	while (sl_strlen(key) < KEY_LEN ) { 
	  if (key[0] != '\n' && key[0] != '\0')
	    fprintf(stdout, _("New audit trail, enter key: "));
	  fflush(stdout); 
	  key[0] = '\0';
	  fgets(key, KEY_LEN+1, stdin); 
	}
	sh_util_encode(key, buf, 1);
	start = 0;
      } 
    else               /* iterate key */
      {             
	sl_strlcpy (key, 
		    sh_tiger_hash (key, TIGER_DATA, KEY_LEN), 
		    KEY_LEN+1);
      }

    sl_strlcat ( buf, key, 2*SH_BUFSIZE + 1);


#ifdef SH_STEALTH
    sh_do_decode (signature, sl_strlen(signature));
#endif


    if (sl_strncmp (signature, 
		    sh_tiger_hash (buf, TIGER_DATA, sl_strlen(buf)),
		    KEY_LEN) != 0) {
      buf[len] = '\0';    /* do not print out the key */
#ifdef SH_STEALTH
      sh_do_decode (buf, sl_strlen(buf));
#endif
      fprintf (stdout, _("FAILED (%5d): %s\n"), count-1, buf);

    } else { 
      buf[len] = '\0';    /* do not print out the key */
#ifdef SH_STEALTH
      sh_do_decode (buf, sl_strlen(buf));
#endif
      fprintf (stdout, _("passed (%5d): %s\n"), count-1, buf);
    }
  }

  sl_close (fd);
  SH_FREE(buf);

  fflush(stdout);

  _exit (EXIT_SUCCESS);
  return 0;
}

/*********************************************
 *  end utility functions
 *********************************************/

void sh_error_only_stderr (int flag)
{
  OnlyStderr    = flag;
  return;
}

static int dbg_store = 0;
static int dbg_flag  = 0;

void sh_error_dbg_on()
{
  if (dbg_flag == 1)
    return;

  dbg_store           = errFlags.printlevel;
  errFlags.printlevel = (SH_ERR_ALL    | SH_ERR_INFO  | SH_ERR_NOTICE | 
			 SH_ERR_WARN   | SH_ERR_STAMP | SH_ERR_ERR    | 
			 SH_ERR_SEVERE | SH_ERR_FATAL);
  dbg_flag  = 1;
  return;
}

void sh_error_dbg_off()
{
  if (dbg_flag == 0)
    return;

  errFlags.printlevel = dbg_store;
  dbg_store = 0;
  dbg_flag  = 0;
  return;
}



char * sh_error_message (int tellme)
{

#if defined(HAVE_STRERROR)
  return strerror(tellme);
#else

#ifdef EACCES
    if (tellme == EACCES)  return _("Permission denied.");
#endif
#ifdef EAGAIN
    if (tellme == EAGAIN)  return _("Try again.");
#endif
#ifdef EBADF
    if (tellme == EBADF)   return _("File descriptor in bad state.");
#endif
#ifdef EEXIST
    if (tellme == EEXIST)  return _("File exists.");
#endif
#ifdef EFAULT
    if (tellme == EFAULT)  return _("Bad address.");
#endif
#ifdef EINVAL
    if (tellme == EINVAL)  return _("Invalid argument.");
#endif
#ifdef EISDIR
    if (tellme == EISDIR)  return _("Is a directory.");
#endif
#ifdef EINTR
    if (tellme == EINTR)   return _("System call was interrupted.");
#endif
#ifdef EIO
    if (tellme == EIO)     return _("Low-level I/O error.");
#endif
#ifdef ELOOP
    if (tellme == ELOOP)   return _("Too many symbolic links encountered.");
#endif
#ifdef EMFILE
    if (tellme == EMFILE)  return _("Too many open files.");
#endif
#ifdef EMLINK
    if (tellme == EMLINK)  return _("Too many links.");
#endif
#ifdef ENAMETOOLONG
    if (tellme == ENAMETOOLONG) 
                           return _("File name too long."); 
#endif
#ifdef ENFILE
    if (tellme == ENFILE)  return _("File table overflow.");
#endif
#ifdef ENOENT
    if (tellme == ENOENT)  return _("File does not exist.");
#endif
#ifdef ENOMEM
    if (tellme == ENOMEM)  return _("Out of memory.");
#endif
#ifdef ENOSPC
    if (tellme == ENOSPC)  return _("No space on device.");
#endif
#ifdef ENOTDIR
    if (tellme == ENOTDIR) return _("Not a directory.");
#endif
#ifdef ENOTSOCK
    if (tellme == ENOTSOCK) return _("Not a socket.");
#endif
#ifdef EOPNOTSUPP
    if (tellme == EOPNOTSUPP) return _("Socket is not of type SOCK_STREAM.");
#endif
#ifdef EPERM
    if (tellme == EPERM)   return _("Permission denied.");
#endif
#ifdef EPIPE
    if (tellme == EPIPE)   return _("No read on pipe.");
#endif
#ifdef EROFS
    if (tellme == EROFS)    return _("Read-only file system.");
#endif
#ifdef ETXTBSY
    if (tellme == ETXTBSY) return _("Text file busy.");
#endif
#ifdef EWOULDBLOCK
    if (tellme == EWOULDBLOCK) 
      return _("No connections on non-blocking socket.");
#endif
#ifdef EXDEV
    if (tellme == EXDEV)    return _("Not on same file system.");
#endif
    return _("Unknown error");
#endif /* ifndef HAVE_STRERROR */
}


/* switch off file log
 */
void sh_error_logoff()
{
  errFlags.HaveLog = BAD;
  return;
}

/* switch on file log 
 */
void sh_error_logrestore()
{
  errFlags.HaveLog = GOOD;
  return;
}

typedef struct eef 
{
  char * str;
  int    val;
} eef_struc;

static eef_struc eef_tab[] =
{
  { N_("none"),    SH_ERR_NOT    },
  { N_("debug"),   SH_ERR_ALL    },
  { N_("info"),    SH_ERR_INFO   },
  { N_("notice"),  SH_ERR_NOTICE },
  { N_("warn"),    SH_ERR_WARN   },
  { N_("mark"),    SH_ERR_STAMP  },
  { N_("err"),     SH_ERR_ERR    },
  { N_("crit"),    SH_ERR_SEVERE },
  { N_("alert"),   SH_ERR_FATAL  },
};
#define SH_EEF_MAX 9

/* set severity levels
 */
int sh_error_set_iv (int iv, char *  str_s)
{
  int level = (-1);
  int i;

  for (i = 0; i < SH_EEF_MAX; ++i)
    {
      if (0 == sl_strncmp(str_s, _(eef_tab[i].str), 
		       sl_strlen(eef_tab[i].str))) 
	{
	  level = eef_tab[i].val;
	  break;
	}
    }

  if (level == (-1)) 
    {
      sh_error_handle (SH_ERR_ERR,
		       _("Set severity - unknown severity %s"),
		       str_s);
      return (-1);
    }

  if (iv > SH_ERR_T_START && iv < SH_ERR_T_END) 
    {
      ShDFLevel[iv] =  level;
    } 
  else 
    {
      sh_error_handle (SH_ERR_ERR, 
		       _("Set severity - unknown policy %d"),
		       (int) iv);
      return (-1);
    }
  return (0);
}

int sh_error_set_level(char * str_s, int *facility)
{
  register int  i, j, f = BAD;

  int  old_facility;

  if (IsInitialized == BAD) 
    (void) sh_error_init();

  old_facility = *facility;
  *facility    = 0;

  if (str_s != NULL) 
    {
      if (0 == sl_strncmp(str_s, _(eef_tab[0].str), sl_strlen(eef_tab[0].str)))
	{
	  *facility |= eef_tab[0].val;
	  f = GOOD;
	}
      else if (str_s[0] == '*')
	{
	  for (i = 1; i < SH_EEF_MAX; ++i)
	    *facility |= eef_tab[i].val;
	  f = GOOD;
	}
      else if (str_s[0] == '=')
	{
	  for (i = 1; i < SH_EEF_MAX; ++i)
	    if (0 == sl_strncmp(&str_s[1], _(eef_tab[i].str), 
			     sl_strlen(eef_tab[i].str)))
		{ *facility |= eef_tab[i].val; f = GOOD; }
	}
      else if (str_s[0] == '!')
	{
	  for (i = 1; i < SH_EEF_MAX; ++i)
	    if (0 != sl_strncmp(&str_s[1], _(eef_tab[i].str), 
			     sl_strlen(eef_tab[i].str)))
		{ *facility |= eef_tab[i].val;  f = GOOD; }
	}
      else
	{
	  for (i = 1; i < SH_EEF_MAX; ++i)
	    {
	      if (0 == sl_strncmp(str_s, _(eef_tab[i].str), 
			       sl_strlen(eef_tab[i].str))) 
		{
		  for (j = i; j < SH_EEF_MAX; ++j)
		    {
		      *facility |= eef_tab[j].val;
		    }
		  f = GOOD; 
		  break;
		}
	    }
	}
    }

  if (f == BAD) 
    {
      *facility = old_facility; 
      sh_error_handle (SH_ERR_ERR, 
		       _("Bad log priority: %s"), str_s);
      return (-1);
    }


  return (0);
}

#ifdef SH_WITH_CLIENT
/* set severity for TCP export
 */
int sh_error_setexport(char *  str_s)
{
  static int reject = 0;
  if (reject == 1)
    return (-1);

  if (sh.opts == 1)  
    reject = 1;

  return (sh_error_set_level(str_s, &errFlags.exportlevel));
}
#endif

/* set severity for printing
 */
int sh_error_setprint(char *  str_s)
{
  static int reject = 0;
  if (reject == 1)
    return (-1);

  if (sh.opts == 1)  
    reject = 1;

  return (sh_error_set_level(str_s, &errFlags.printlevel));
}


/* set level for error logging
 */
int sh_error_setlog(char * str_s)
{
  static int reject = 0;
  if (reject == 1)
    return (-1);

  if (sh.opts == 1)  
    reject = 1;

  return ( sh_error_set_level(str_s, &errFlags.loglevel) );
}


/* set severity for syslog
 */
int sh_error_set_syslog (char * str_s)
{
  static int reject = 0;
  if (reject == 1)
    return (-1);

  if (sh.opts == 1)  
    reject = 1;

  return (sh_error_set_level(str_s, &errFlags.sysloglevel));
}

#if defined(WITH_EXTERNAL)
/* set severity for external
 */
int sh_error_set_external (char * str_s)
{
  static int reject = 0;
  if (reject == 1)
    return (-1);

  if (sh.opts == 1)  
    reject = 1;

  return (sh_error_set_level(str_s, &errFlags.externallevel));
}
#endif

/* set severity for mailing
 */
int sh_error_setseverity (char * str_s)
{
  static int reject = 0;
  if (reject == 1)
    return (-1);

  if (sh.opts == 1)  
    reject = 1;

  return (sh_error_set_level(str_s, &errFlags.maillevel));
}

  
#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
#include <sys/mman.h>
#endif


/**********************************************************
 **********************************************************
 *
 * --------  MAIN ERROR HANDLING FUNCTION -----------------
 *
 *
 * this function should be called to report an error
 *
 ********************************************************** 
 **********************************************************/
void sh_error_handle (ShErrLevel  severity, 
		      const char *fmt, ...)
{
  va_list       vl;                 /* argument list      */
  char        * message = NULL;     /* pointer to message */

  /* block a facility for errors generated
   * within that facility
   */
  static int print_block  = 0;
  static int mail_block   = 0;
  static int syslog_block = 0;
  static int log_block    = 0;
#ifdef SH_WITH_CLIENT
  static int export_block = 0;
#endif
#if defined(WITH_EXTERNAL)
  static int external_block = 0;
#endif


  /* initialize to default values
   */
  if (IsInitialized == BAD) 
    sh_error_init();


  /* consistency checks
   */
  if ((int)(severity) > (int)(SH_ERR_MAX) || 
      (int)(severity) < (int)(SH_ERR_ALL)) 
    {
      sh_error_handle (SH_ERR_ERR,
		       _("Internal error: Invalid severity in error message"));
      return;
    }
  else if (fmt == NULL) 
    {
      sh_error_handle (SH_ERR_ERR,
		       _("Internal error: fmt in error message is NULL"));
      return;
    }

  /* messages not wanted for logging
   */
  else if (
	   (
	    (errFlags.printlevel  & severity) == 0    &&
	    (errFlags.loglevel    & severity) == 0    &&
	    (errFlags.sysloglevel & severity) == 0    &&
#ifdef SH_WITH_CLIENT
	    (errFlags.exportlevel & severity) == 0    &&
#endif
#ifdef WITH_EXTERNAL
	    (errFlags.externallevel & severity) == 0  &&
#endif
	    (errFlags.maillevel   & severity) == 0   
#ifdef SH_WITH_SERVER
	    && (severity  & SH_ERR_INET) == 0
#endif
	    ) || (severity & SH_ERR_NOT) != 0
	   )
    return;


  message = SH_ALLOC(SH_BUFSIZE+2);
  MLOCK( message, SH_BUFSIZE+2);
  message[0] = '\0';

  /* format the log message with timestamp etc
   */
  va_start (vl, fmt);
  sh_error_string (message, severity, fmt, vl);
  va_end (vl);


  /* log to stderr
   */
  if ( ((errFlags.printlevel & severity) != 0 && 
      (errFlags.printlevel & SH_ERR_NOT) == 0)
#ifdef SH_WITH_SERVER
      || (severity  & SH_ERR_INET) != 0
#endif
      )
    { 
      if (print_block == 0 && (errFlags.printlevel & SH_ERR_NOT) == 0) 
	{
	  print_block = 1;
	  sh_error_print (message);
	  print_block = 0;
	}
    }

  if (OnlyStderr == BAD)  /* full error logging enabled */
    {

      /* log to syslog
       */
      if ( (errFlags.sysloglevel & severity) != 0 && 
	   (severity & SH_ERR_INET) == 0  && /* don't log inet to syslog */
	   (errFlags.sysloglevel & SH_ERR_NOT) == 0 ) 
	{
	  if (syslog_block == 0)
	    {
	      syslog_block = 1;
	      sh_error_syslog (severity, message);
	      syslog_block = 0;
	    }
	}

#if defined(WITH_EXTERNAL)
      /* 
       * -- external facility 
       */
      if ((errFlags.externallevel & severity) != 0 && 
	  (errFlags.externallevel & SH_ERR_NOT) == 0  )
	{
	  if (external_block == 0)
	    {
	      external_block = 1;
	      sh_ext_execute ('l', 'o', 'g', message, 0);
	      external_block = 0;
	    }
	}
#endif

      /*****************
       * optionally include client code for TCP forwarding
       * to log server
       *****************/
#ifdef SH_WITH_CLIENT
      /* export by TCP
       */
      if ((errFlags.exportlevel & severity) != 0 &&
          (errFlags.exportlevel & SH_ERR_NOT) == 0 &&
          sh.isserver != GOOD                            &&
          (severity & SH_ERR_INET) == 0 ) /* don't log inet to export */
        {
          if (export_block == 0)
            {
              export_block = 1;
              sh_forward (message);
              export_block = 0;
            }
        }
#endif


      /* log to mail
       */
      if ((errFlags.maillevel & severity) != 0 &&
	  (errFlags.maillevel & SH_ERR_NOT) == 0 &&
	  (severity & SH_ERR_INET) == 0 ) /* don't log inet to email */
	{
	  if (mail_block == 0)
	    {
	      mail_block = 1;
	      if ( (severity & SH_ERR_FATAL) == 0) {
		sh_mail_pushstack (message);
	      } else {
		sh_mail_msg (sh.address, message);
	      }
	      mail_block = 0;
	    }
	}

      /* log to logfile
       */

      if ( ( (  (errFlags.loglevel & severity) != 0 &&
	      (errFlags.loglevel & SH_ERR_NOT) == 0 )
#ifdef SH_WITH_SERVER
	     || (severity  & SH_ERR_INET) != 0
#endif
	     ) &&
	  (errFlags.HaveLog != BAD) &&  /* temporary switched off */
	  (severity & SH_ERR_NOT) == 0 /* paranoia */
	  ) 
	{
	  if (log_block == 0)
	    {
	      log_block = 1;
	      sh_error_log (message);
	      log_block = 0;
	    }
	}

    }

  /* unlock
   */
  memset ( message, '\0', SH_BUFSIZE+2 );
  MUNLOCK( message,      SH_BUFSIZE+2 );
  SH_FREE(message);

  return;
}


/* -------------------------  
 *
 * private functions below
 *
 * -------------------------
 */

/* print an error  into string
 */
static int sh_error_string (char *errmsg, ShErrLevel severity,
			    const char *fmt, va_list vl)
{
  static char   errhead[81+17+9];
  static char   errshort[64+2];
  char        * errbody;
  int           isshort = 0;

  /* header of error message
   */
  if      ( (severity & SH_ERR_INET) != 0)
    sl_strlcpy (errhead, _("<TCP>  : "), 11);
  else if ( (severity & SH_ERR_ALL) != 0)
    sl_strlcpy (errhead, _("DEBUG  : "), 11);
  else if ( (severity & SH_ERR_INFO) != 0)
    sl_strlcpy (errhead, _("INFO   : "), 11);
  else if ( (severity & SH_ERR_NOTICE) != 0)
    sl_strlcpy (errhead, _("NOTICE : "), 11);
  else if ( (severity & SH_ERR_WARN) != 0)
    sl_strlcpy (errhead, _("WARN   : "), 11);
  else if ( (severity & SH_ERR_STAMP) != 0)
    sl_strlcpy (errhead, _("MARK   : "), 11);
  else if ( (severity & SH_ERR_ERR) != 0)
    sl_strlcpy (errhead, _("ERROR  : "), 11);
  else if ( (severity & SH_ERR_SEVERE) != 0)
    sl_strlcpy (errhead, _("CRIT   : "), 11);
  else if ( (severity & SH_ERR_FATAL) != 0)
    sl_strlcpy (errhead, _("ALERT  : "), 11);
  else {
    sl_strlcpy (errhead, _("???    : "), 11);
  }

  /* body of error message -- do not use VA_COPY if not required,
   */
  if ( (NULL == strchr(fmt, '%')) && (sl_strlen(fmt) <= 64)) 
    {
      isshort = 1;
      sl_strlcpy (errshort, fmt, sizeof(errshort));
      errbody = &errshort[0];
      MLOCK( errbody, sizeof(errshort));
    }

  /* use VA_COPY 
   */
  else 
    {
      isshort = 0;
      errbody = SH_ALLOC(SH_BUFSIZE+2);
      MLOCK( errbody, SH_BUFSIZE+2);
      sl_vsnprintf(errbody, (SH_BUFSIZE-sl_strlen(errhead)), fmt, vl);
    }

  if (severity != SH_ERR_ALL)  
    {
      sl_strlcpy (errmsg, errhead,          SH_BUFSIZE+1);
      sl_strlcat (errmsg, sh_unix_time (0), SH_BUFSIZE+1);
      sl_strlcat (errmsg, " ",              SH_BUFSIZE+1);
      sl_strlcat (errmsg, errbody,          SH_BUFSIZE+1);
    }
  else 
    {  
      sl_strlcpy (errmsg, errhead,          SH_BUFSIZE+1);
      sl_strlcat (errmsg, errbody,          SH_BUFSIZE+1);
    }

  if (isshort == 0) {
    memset( errbody, '\0', SH_BUFSIZE+2);
    MUNLOCK( errbody, SH_BUFSIZE+2);
    SH_FREE(errbody);
  } else {
    memset( errbody, '\0', sizeof(errshort));
    MUNLOCK( errbody, sizeof(errshort));
  }

  return(0);
}

     
/* syslog error message
 */
static int  sh_error_syslog (ShErrLevel  severity, char *errmsg)
{
  int priority;
  static int init = 0;

  if      (severity == SH_ERR_ALL)    priority = LOG_DEBUG;
  else if (severity == SH_ERR_INFO)   priority = LOG_INFO;
  else if (severity == SH_ERR_NOTICE) priority = LOG_NOTICE;
  else if (severity == SH_ERR_WARN)   priority = LOG_WARNING;
  else if (severity == SH_ERR_STAMP)  priority = LOG_ERR;
  else if (severity == SH_ERR_ERR)    priority = LOG_ERR;
  else if (severity == SH_ERR_SEVERE) priority = LOG_CRIT;
  else if (severity == SH_ERR_FATAL)  priority = LOG_ALERT;
  else priority = LOG_DEBUG;

  /* limit message size
   */
  if (sl_strlen(errmsg) > 1023) 
    errmsg[1023] = '\0';

  if (init == 0)
    {
#ifdef LOG_AUTHPRIV
      openlog (_("samhain"), LOG_PID, LOG_AUTHPRIV);
#else
      openlog (_("samhain"), LOG_PID, LOG_AUTH);
#endif
      init = 1;
    }
  syslog (priority, "%s", errmsg);       /* -- errmsg is < 1024 chars -- */
  return (0);
}

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* print out error message
 */
static int  sh_error_print (char *errmsg)
{
  int fd;
  FILE * ofile;

  if (sh.modus == 0) /* not daemon */ 
    {
      fprintf (stderr, "%s\n", errmsg);
    }
  else if (OnlyStderr == BAD) /* daemon && initialized */
    {
      fd    =   open (_PATH_CONSOLE, O_WRONLY|O_APPEND|O_NOCTTY);
      if (fd > 0)
	{
	  ofile = fdopen (fd, "a");
	  if (ofile != NULL)
	    {
	      (void) fprintf (ofile, "%s\r\n", errmsg);
	      (void) fclose  (ofile);
	    }
	}
    }
  return (0);
}

/* log error message
 */
static int  sh_error_log (char *errmsg)
{
  ShErrLevel    flag;
  ShErrLevel    store;
  SL_TICKET     fd;
  char        * tmp;
  /* static int    isfirst = 1; */
  char          signature[61];
  char          buf[SH_BUFSIZE + 53 + KEY_LEN + 18];
  

  if (errFlags.HaveLog == BAD)  /* paranoia */ 
    return (-1);

  /* open/create the file, then check it 
   */


  if (  0 !=  tf_trust_check (DEFAULT_ERRFILE, SL_YESPRIV)          ||
      SL_ISERROR(fd = sl_open_write (errFlags.errfile, SL_YESPRIV)) ||
      sh_unix_testlock(DEFAULT_ERRFILE, SH_ERR_FATAL) < 0           ||
      SL_ISERROR(sl_forward(fd) )) {
    sh_error_logoff();
    tmp  = sh_util_safe_name (errFlags.errfile);
    if (sh.locked == BAD)
      sh_error_handle (SH_ERR_FATAL,
		       _("PANIC -- could not open logfile %s: file is locked"),
		       tmp);
    else
      sh_error_handle (SH_ERR_FATAL,
		       _("PANIC -- could not open logfile %s"),
		       tmp);
    SH_FREE(tmp);
    return (-1);
  }

#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
  if (skey->mlock_failed == GOOD) 
    {
      sl_set_suid ();
      if ( (-1) == mlock( buf, sizeof(buf)) ) 
	{
	  sl_unset_suid ();
	  skey->mlock_failed = BAD;
	  sh_error_handle (SH_ERR_WARN,
			   _("Using insecure memory"));
	}
      else
	sl_unset_suid ();
    }
#else
    if (skey->mlock_failed == GOOD) {
      skey->mlock_failed = BAD;
      sh_error_handle (SH_ERR_WARN,
		       _("Using insecure memory"));
    }
#endif

  /* write the start marker 
   */
  if (sh.isfirst == 1) {
#ifdef SH_STEALTH
    sl_write (fd, "\n", 1);
    sl_write_line (fd, N_("[SOF]"), 5);
    sl_sync(fd);
#else
    sl_write_line (fd, _("\n[SOF]"),  6);
    sl_sync(fd);
#endif
  }

  /* reserve KEY_LEN chars at end for key 
   */
  sl_strlcpy (buf, errmsg, sizeof(buf)-KEY_LEN);

#ifdef SH_STEALTH
  sh_do_encode (buf, sl_strlen(buf));
#endif

  /* write() OUT 
   */
  sl_write_line (fd, buf, sl_strlen (buf));
  sl_sync(fd);
  

  /* write the signature 
   */
  if (sh.isfirst == 1) {

    store                = errFlags.loglevel;
    flag                 = errFlags.sysloglevel;
    errFlags.loglevel    = SH_ERR_NOT;
    errFlags.sysloglevel = SH_ERR_NOT;

    if (sh.real_user[0] == '\0') (void) sh_unix_getUser();

    sh_util_keyinit(skey->sigkey_old, KEY_LEN+1); 
    sl_strlcpy(skey->sigkey_new, 
	    sh_tiger_hash (skey->sigkey_old, TIGER_DATA, KEY_LEN), 
	    KEY_LEN+1);
    sl_strlcpy(skey->crypt, skey->sigkey_new, KEY_LEN+1);
    sh_util_encode(skey->crypt, buf, 0);
    sh_error_handle (SH_ERR_FATAL,
#if defined(SH_WITH_SERVER) && !defined(SH_WITH_CLIENT)
		     _("LOG_KEY yule    %s"),
#else
		     _("LOG_KEY samhain %s"),
#endif
		     skey->crypt);
    memset (skey->crypt, '\0', KEY_LEN);
    errFlags.sysloglevel = flag;
    errFlags.loglevel    = store;

    sh.isfirst = 0;  /* signature key is initialized */
	
  } else {

    sl_strlcpy (skey->sigkey_new, 
	     sh_tiger_hash (skey->sigkey_old, TIGER_DATA, KEY_LEN),
	     KEY_LEN+1);
    
  }

  /* sign the message with the signature key
   */
  sl_strlcat (buf, skey->sigkey_new, sizeof(buf));
  
  sl_strlcpy (signature,
	      sh_tiger_hash (buf, TIGER_DATA, sl_strlen(buf)), 
	      sizeof(signature));
  sl_strlcpy (skey->sigkey_old, skey->sigkey_new, KEY_LEN+1); 


#ifdef SH_STEALTH
  sh_do_encode (signature, sl_strlen(signature));
#endif

  sl_write_line (fd, signature, sl_strlen (signature));
  sl_sync (fd);
  sl_close (fd);

  memset (buf, '\0', sizeof(buf));
  MUNLOCK( buf, sizeof(buf) );


  return (0);
}


/* initialize
 */
static int  sh_error_init ()
{
  register int j;

  errFlags.debug          = 0;
  errFlags.HaveLog        = GOOD;
  errFlags.sysloglevel    = SH_ERR_NOT;
  errFlags.loglevel       = SH_ERR_NOT;
  errFlags.externallevel  = SH_ERR_NOT;
  errFlags.maillevel      = SH_ERR_FATAL;
#if defined(SH_STEALTH)
  errFlags.printlevel     = SH_ERR_NOT;
#else
  errFlags.printlevel     = (SH_ERR_INFO  | SH_ERR_NOTICE | SH_ERR_WARN   | 
			     SH_ERR_STAMP | SH_ERR_ERR    | SH_ERR_SEVERE |
			     SH_ERR_FATAL);
#endif
  errFlags.exportlevel    = (SH_ERR_STAMP | SH_ERR_ERR    | SH_ERR_SEVERE |
			     SH_ERR_FATAL);
  sl_strlcpy(errFlags.errfile, DEFAULT_ERRFILE, PATH_MAX+1);

  for (j = 0; j < 12; ++j) {
    ShDFLevel[j] = SH_ERR_NOT;
  }

  IsInitialized = GOOD;
  return (0);
}

