/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include "config_xor.h"

#ifdef HAVE_BROKEN_INCLUDES
#define _ANSI_C_SOURCE
#define _POSIX_SOURCE
#define _XOPEN_SOURCE_EXTENDED
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#ifdef  HAVE_UNISTD_H
#include <errno.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <unistd.h>

/*********************
#ifdef HAVE_SYS_VFS_H
#include <sys/vfs.h>
#endif
**********************/
#endif

#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifndef FD_SET
#define NFDBITS         32
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#endif /* !FD_SET */
#ifndef FD_SETSIZE
#define FD_SETSIZE      32
#endif
#ifndef FD_ZERO
#define FD_ZERO(p)      memset((char *)(p), '\0', sizeof(*(p)))
#endif


#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
#include <sys/mman.h>
#endif

#include "samhain.h"
#include "sh_error.h"
#include "sh_unix.h"
#include "sh_utils.h"
#include "sh_mem.h"
#include "sh_hash.h"
#include "sh_tiger.h"
#include "sh_tools.h"


#ifndef HAVE_LSTAT
#define lstat   stat
#endif 


extern char **environ;

#ifdef SYS_SIGLIST_DECLARED

extern const char * const sys_siglist[];

#else

char * sh_unix_siglist (int signum)
{

  switch (signum)
    {
#ifdef SIGHUP
    case SIGHUP: 
      return _("Hangup");
#endif
#ifdef SIGINT
    case SIGINT: 
      return _("Interrupt");
#endif
#ifdef SIGQUIT
    case SIGQUIT: 
      return _("Quit");
#endif
#ifdef SIGILL
    case SIGILL: 
      return _("Illegal instruction");
#endif
#ifdef SIGTRAP
    case SIGTRAP: 
      return _("Trace/breakpoint trap");
#endif
#ifdef SIGABRT
    case SIGABRT: 
      return _("IOT trap/Abort");
#endif
#ifdef SIGBUS
    case SIGBUS: 
      return _("Bus error");
#endif
#ifdef SIGFPE
    case SIGFPE: 
      return _("Floating point exception");
#endif
#ifdef SIGUSR1
    case SIGUSR1: 
      return _("User defined signal 1");
#endif
#ifdef SIGSEGV
    case SIGSEGV: 
      return _("Segmentation fault");
#endif
#ifdef SIGUSR2
    case SIGUSR2: 
      return _("User defined signal 2");
#endif
#ifdef SIGPIPE
    case SIGPIPE: 
      return _("Broken pipe");
#endif
#ifdef SIGALRM
    case SIGALRM: 
      return _("Alarm clock");
#endif
#ifdef SIGTERM
    case SIGTERM: 
      return _("Terminated");
#endif
#ifdef SIGSTKFLT
    case SIGSTKFLT: 
      return _("Stack fault");
#endif
#ifdef SIGCHLD
    case SIGCHLD: 
      return _("Child exited");
#endif
#ifdef SIGCONT
    case SIGCONT: 
      return _("Continued");
#endif
#ifdef SIGSTOP
    case SIGSTOP: 
      return _("Stopped");
#endif
#ifdef SIGTSTP
    case SIGTSTP: 
      return _("Stopped");
#endif
#ifdef SIGTTIN
    case SIGTTIN: 
      return _("Stopped (tty input)");
#endif
#ifdef SIGTTOU
    case SIGTTOU: 
      return _("Stopped (tty output)");
#endif
#ifdef SIGURG
    case SIGURG: 
      return _("Urgent condition");
#endif
#ifdef SIGXCPU
    case SIGXCPU: 
      return _("CPU time limit exceeded");
#endif
#ifdef SIGXFSZ
    case SIGXFSZ: 
      return _("File size limit exceeded");
#endif
#ifdef SIGVTALRM
    case SIGVTALRM: 
      return _("Virtual time alarm");
#endif
#ifdef SIGPROF
    case SIGPROF: 
      return _("Profile signal");
#endif
#ifdef SIGWINCH
    case SIGWINCH: 
      return _("Window size changed");
#endif
#ifdef SIGIO
    case SIGIO: 
      return _("Possible I/O");
#endif
#ifdef SIGPWR
    case SIGPWR: 
      return _("Power failure");
#endif
#ifdef SIGUNUSED
    case SIGUNUSED: 
      return _("Unused signal");
#endif
    }
  return _("?");
}

#endif

extern char sh_sig_msg[64];

static
void sh_unix_sigexit (int signal
#if defined(SA_SIGINFO) && defined(SI_USER)
		      , siginfo_t * signal_info, void * signal_add
#endif
		      )
{
#if defined(SA_SIGINFO) && defined(SI_USER) && !defined(__linux__)
  /* On linux (as of 2.2.16) no useful information in 
   * siginfo_t * signal_info.
   * On Solaris, signal_info may be a NULL pointer :-(
   */
  if (signal_info != NULL && signal_info->si_code == SI_USER && 
      signal != SIGTERM)
    return;
#endif

  strncpy (sh_sig_msg, _("on signal "), 20);

  sh_sig_msg[10] = '\0';

#ifdef SYS_SIGLIST_DECLARED
  strncat (sh_sig_msg, sys_siglist[signal],     40);
#else
  strncat (sh_sig_msg, sh_unix_siglist(signal), 40);
#endif

  sh_sig_msg[63] = '\0';

  exit(EXIT_SUCCESS);
}

volatile int immediate_exit_fast = 0;

static
void sh_unix_sigexit_fast (int signal
#if defined(SA_SIGINFO) && defined(SI_USER)
			   , siginfo_t * signal_info, void * signal_add
#endif
			   )
{
  /* On Linux there is no useful information in siginfo_t
   */
#if defined(SA_SIGINFO) && defined(SI_USER) && !defined(__linux__)
  /* On linux (as of 2.2.16) no useful information in 
   * siginfo_t * signal_info.
   * On Solaris, signal_info may be a NULL pointer :-(
   */
  if (signal_info != NULL && signal_info->si_code == SI_USER)
    return;
#endif


  if (immediate_exit_fast == 1)
    {
      ++immediate_exit_fast;
      memset (skey, '\0', sizeof(sh_key_t));
      _exit(EXIT_FAILURE);
    }
  else if (immediate_exit_fast != 0)
    _exit(EXIT_FAILURE);

  ++immediate_exit_fast;
  
  /* the following signals leave the system in an undefined
   * state, thus it is best to exit immediately
   */
#ifdef SIGSEGV
  if      (signal == SIGSEGV)
    sh_error_handle (SH_ERR_FATAL,
		     _("EXIT on invalid memory reference."));
#endif
#ifdef SIGILL
  if (signal == SIGILL)
    sh_error_handle (SH_ERR_FATAL,
		     _("EXIT on illegal instruction."));
#endif
#ifdef SIGBUS
  if (signal == SIGBUS)
    sh_error_handle (SH_ERR_FATAL,
		     _("EXIT on bus error."));
#endif
#ifdef SIGFPE
  if (signal == SIGFPE)
    sh_error_handle (SH_ERR_FATAL,
		     _("EXIT on floating point exception."));
#endif

  memset (skey, '\0', sizeof(sh_key_t));
  _exit(EXIT_FAILURE); 
}


static
void sh_unix_sigaction (int signal)
{
  sig_raised = 1;
#ifdef SIGUSR1
  if (signal == SIGUSR1)
    sig_debug_on          = 1;
#endif
#ifdef SIGUSR2
  if (signal == SIGUSR2)
    sig_debug_off         = 1;
#endif
#ifdef SIGHUP
  if (signal == SIGHUP)
    sig_config_read_again = 1;
#endif
#ifdef SIGABRT
  if (signal == SIGABRT)
    sig_fresh_trail       = 1;
#endif
#ifdef SIGQUIT
  if (signal == SIGQUIT)
    sig_terminate         = 1;
#endif

  return;
}


void sh_unix_siginstall ()
{
  struct sigaction act, act_fast, act2, oldact, ignact;


  ignact.sa_handler = SIG_IGN;          /* signal action           */
  sigemptyset( &ignact.sa_mask );       /* set an empty mask       */
  ignact.sa_flags = 0;                  /* init sa_flags           */

#if defined(SA_SIGINFO) && defined(SI_USER)
  act.sa_sigaction = &sh_unix_sigexit;    /* signal action           */
#else
  act.sa_handler   = &sh_unix_sigexit;    /* signal action           */
#endif
  sigfillset ( &act.sa_mask );            /* set a  full mask        */
#ifdef SIGILL
  sigdelset  ( &act.sa_mask, SIGILL  );
#endif
#ifdef SIGFPE
  sigdelset  ( &act.sa_mask, SIGFPE  );
#endif
#ifdef SIGSEGV
  sigdelset  ( &act.sa_mask, SIGSEGV );
#endif
#ifdef SIGBUS
  sigdelset  ( &act.sa_mask, SIGBUS  );
#endif

#if defined(SA_SIGINFO) && defined(SI_USER)
  act_fast.sa_sigaction = &sh_unix_sigexit_fast;  /* signal action           */
#else
  act_fast.sa_handler   = &sh_unix_sigexit_fast;  /* signal action           */
#endif
  sigfillset ( &act_fast.sa_mask );               /* set a full mask         */
#ifdef SIGILL
  sigdelset  ( &act_fast.sa_mask, SIGILL  );
#endif
#ifdef SIGFPE
  sigdelset  ( &act_fast.sa_mask, SIGFPE  );
#endif
#ifdef SIGSEGV
  sigdelset  ( &act_fast.sa_mask, SIGSEGV );
#endif
#ifdef SIGBUS
  sigdelset  ( &act_fast.sa_mask, SIGBUS  );
#endif


  /* fallback to default action if sh_unix_sigexit fails for
   * some reason (e.g. segfault in exit)
   */
#if defined(SA_SIGINFO) && defined(SI_USER)
  act.sa_flags      = SA_SIGINFO;
  act_fast.sa_flags = SA_SIGINFO;
#else
  act.sa_flags      = 0;
  act_fast.sa_flags = 0;
#endif 

#if   defined(SA_NOMASK)
  act_fast.sa_flags |= SA_NOMASK;
#elif defined(SA_NODEFER)
  act_fast.sa_flags |= SA_NODEFER;
#endif

#if 0
#if defined(SA_RESETHAND)
  act.sa_flags = SA_RESETHAND 
#if defined(SA_SIGINFO) && defined(SI_USER)
    | SA_SIGINFO
#endif
    ;          
  act_fast.sa_flags = SA_RESETHAND     
#if defined(SA_SIGINFO) && defined(SI_USER)
    | SA_SIGINFO
#endif
    ;          
#elif defined(SA_ONESHOT)
  act.sa_flags = SA_ONESHOT            
#if defined(SA_SIGINFO) && defined(SI_USER)
    | SA_SIGINFO
#endif
    ;          
  act_fast.sa_flags = SA_ONESHOT       
#if defined(SA_SIGINFO) && defined(SI_USER)
    | SA_SIGINFO
#endif
    ;          
#else
  act.sa_flags = 0
#if defined(SA_SIGINFO) && defined(SI_USER)
    | SA_SIGINFO
#endif
    ;          
  act_fast.sa_flags = 0
#if defined(SA_SIGINFO) && defined(SI_USER)
    | SA_SIGINFO
#endif
    ;          
#endif
#endif


  act2.sa_handler = &sh_unix_sigaction;  /* signal action           */
  sigemptyset( &act2.sa_mask );          /* set an empty mask       */
  act2.sa_flags = 0;                     /* init sa_flags           */


#ifdef SIGHUP
  EINTR_RETRY(sigaction(SIGHUP,  &act2, &oldact));
#endif
#ifdef SIGABRT
  EINTR_RETRY(sigaction(SIGABRT, &act2, &oldact));
#endif
#ifdef SIGUSR1
  EINTR_RETRY(sigaction(SIGUSR1, &act2, &oldact));
#endif
#ifdef SIGUSR2
  EINTR_RETRY(sigaction(SIGUSR2, &act2, &oldact));
#endif
#ifdef SIGQUIT
  EINTR_RETRY(sigaction(SIGQUIT, &act2, &oldact));
#endif
#ifdef SIGTERM
  EINTR_RETRY(sigaction(SIGTERM, &act, &oldact));
#endif

  /* signals that may cause termination */

#ifdef SIGILL
  EINTR_RETRY(sigaction(SIGILL,  &act_fast, &oldact));
#endif
#ifdef SIGFPE
  EINTR_RETRY(sigaction(SIGFPE,  &act_fast, &oldact));
#endif
#ifdef SIGSEGV
  EINTR_RETRY(sigaction(SIGSEGV, &act_fast, &oldact));
#endif
#ifdef SIGBUS
  EINTR_RETRY(sigaction(SIGBUS,  &act_fast, &oldact));
#endif

#ifdef SIGINT
  EINTR_RETRY(sigaction(SIGINT,  &act, &oldact));
#endif
#ifdef SIGPIPE
  EINTR_RETRY(sigaction(SIGPIPE, &act, &oldact));
#endif
#ifdef SIGALRM
  EINTR_RETRY(sigaction(SIGALRM, &ignact, &oldact));
#endif
#ifdef SIGTSTP
  EINTR_RETRY(sigaction(SIGTSTP, &ignact, &oldact));
#endif
#ifdef SIGTTIN
  EINTR_RETRY(sigaction(SIGTTIN, &act, &oldact));
#endif
#ifdef SIGTTOU
  EINTR_RETRY(sigaction(SIGTTOU, &ignact, &oldact));
#endif

#ifdef SIGTRAP
  EINTR_RETRY(sigaction(SIGTRAP, &act, &oldact));
#endif
#ifdef SIGEMT
  EINTR_RETRY(sigaction(SIGEMT,  &ignact, &oldact));
#endif
#ifdef SIGSTKFLT
  EINTR_RETRY(sigaction(SIGSTKFLT,&act, &oldact));
#endif
#ifdef SIGIO
  EINTR_RETRY(sigaction(SIGIO,   &ignact, &oldact));
#endif
#ifdef SIGPOLL
  EINTR_RETRY(sigaction(SIGPOLL, &ignact, &oldact));
#endif
#ifdef SIGXCPU
  EINTR_RETRY(sigaction(SIGXCPU, &act, &oldact));
#endif
#ifdef SIGXFSZ
  EINTR_RETRY(sigaction(SIGXFSZ, &act, &oldact));
#endif
#ifdef SIGVTALRM
  EINTR_RETRY(sigaction(SIGVTALRM, &ignact, &oldact));
#endif
#ifdef SIGPROF
  EINTR_RETRY(sigaction(SIGPROF, &ignact, &oldact));
#endif
#ifdef SIGPWR
  EINTR_RETRY(sigaction(SIGPWR,  &act, &oldact));
#endif

#ifdef SIGLOST
  EINTR_RETRY(sigaction(SIGLOST,  &ignact, &oldact));
#endif
#ifdef SIGUNUSED
  EINTR_RETRY(sigaction(SIGUNUSED,  &ignact, &oldact));
#endif

  return;
}

/* ---------------------------------------------------------------- */

/* checksum the own binary
 */
void sh_unix_self_hash (char * c)
{
  if (c == NULL)
    return;

  sh.sah_path =  SH_ALLOC(sl_strlen(c)+1);
  sl_strlcpy(sh.sah_path, c, sl_strlen(c)+1);

  sl_strlcpy(sh.sah_hash,
	  sh_tiger_hash (c, TIGER_FILE, 0), 
	  KEY_LEN+1);
  if (0 == sl_strcmp(sh.sah_hash,
		  _("0000000000000000000000000000000000000000000000"))
      )
    {
      sh_error_handle (SH_ERR_FATAL, 
		       _("PANIC -- cannot open binary %s"),
		       c);
      exit (EXIT_FAILURE);
    }
  return;
}

int sh_unix_self_check ()
{
  if (sh.sah_path == NULL)
    return (0);
  if (0 == sl_strncmp(sh.sah_hash, 
		   sh_tiger_hash (sh.sah_path, TIGER_FILE, 0), 
		   KEY_LEN))
      return (0);

  sh_error_handle (SH_ERR_FATAL, 
		   _("PANIC -- Binary modified."));
  return (-1);
}


/* ---------------------------------------------------------------- */
  



/* added    Tue Feb 22 10:36:44 NFT 2000 Rainer Wichmann                */
int tf_add_trusted_user(char * c)
{
  register struct passwd *          w;
  static   int             count =  0;
  register long            pwid  = -1;

  if (count == 6)
    {
      fprintf(stderr, _("ERROR  : Maximum 7 trusted users: %s\n"), c);
      return (-1);
    }
  
  /* first check for a user name
   */
  if ((w = getpwnam(c)) != NULL && ((pwid = w->pw_uid) > 0))
    {
      sl_trust_add_user(pwid);
      ++count;
      return (0);
    }
	
  /* failed, so check for a numerical value
   */
  pwid = strtol(c, (char **)NULL, 10);
  if (pwid > 0 && pwid < 65535)
    {
      sl_trust_add_user(pwid);
      ++count;
      return (0);
    }
      
  fprintf(stderr, _("ERROR  : Invalid user: %s\n"), c);
  return (-1);
}

int tf_trust_check (char * file, int mode)
{
  char * tmp;
  char * tmp2;
  int    status;
  int    level;
  uid_t  ff_euid;

  if (mode == SL_YESPRIV)
    sl_get_euid(&ff_euid);
  else
    sl_get_ruid(&ff_euid);

  status = sl_trustfile_euid(file, (long) ff_euid);

  if ( SL_ENONE != status) 
    {
      if (status == SL_ESTAT) 
	level = SH_ERR_ALL;
      else
	level = SH_ERR_ERR;

      tmp  = sh_util_safe_name (file);
      if (sl_trust_errfile != '\0')
	{
	  tmp2  = sh_util_safe_name (sl_trust_errfile());
	  sh_error_handle(level,
			  _("File <%s>: %s At: <%s>."),
			  tmp, sl_error_string(status), tmp2);
	  SH_FREE(tmp2);  
	}
      else
	{
	  sh_error_handle(level,
			  _("File <%s>: %s"),
			  tmp, sl_error_string(status));
	}
      SH_FREE(tmp);

      if (status == SL_EBADUID || status == SL_EBADGID || 
	  status == SL_EBADOTH || status == SL_ETRUNC  || 
	  status == SL_EINTERNAL)
	return (-1);
    }
  return (0);
}

#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 

#if defined(__linux__)

/* determine file type
 */
#include <sys/ioctl.h>
#include <linux/ext2_fs.h>

static 
int sh_unix_getinfo_attr (char * name, 
			  unsigned long * flags, 
			  char * c_attr)
{

/* TAKEN FROM:
 *
 * lsattr.c             - List file attributes on an ext2 file system
 *
 * Copyright (C) 1993, 1994  Remy Card <card@masi.ibp.fr>
 *                           Laboratoire MASI, Institut Blaise Pascal
 *                           Universite Pierre et Marie Curie (Paris VI)
 *
 * This file can be redistributed under the terms of the GNU General
 * Public License
 */

#ifdef HAVE_STAT_FLAGS
  struct stat buf;
  int    i;

  *flags = 0;

  sl_set_suid();                   /* sl_set/unset_suid inserted R.Wichmann */
  i = stat (name, &buf);
  sl_unset_suid();

  if (i == -1)
    return -1;

#ifdef UF_IMMUTABLE
  if (buf.st_flags & UF_IMMUTABLE)
    *flags |= EXT2_IMMUTABLE_FL;
#endif
#ifdef UF_APPEND
  if (buf.st_flags & UF_APPEND)
    *flags |= EXT2_APPEND_FL;
#endif
#ifdef UF_NODUMP
  if (buf.st_flags & UF_NODUMP)
    *flags |= EXT2_NODUMP_FL;
#endif

#else

#ifdef HAVE_EXT2_IOCTLS
  int fd, r, f;
  
  *flags = 0;

  sl_set_suid();                   /* sl_set/unset_suid inserted R.Wichmann */
  fd = open (name, O_RDONLY|O_NONBLOCK);
  sl_unset_suid();

  if (fd == -1)
    return -1;

  r = ioctl (fd, EXT2_IOC_GETFLAGS, &f);
  close (fd);

  if (r == -1)
    return -1;

  *flags = f;
  
#else /* ! HAVE_EXT2_IOCTLS */
  *flags = 0;                                     /* modified by R.Wichmann */
#endif /* ! HAVE_EXT2_IOCTLS */

#endif

/*
 * END
 *
 * lsattr.c             - List file attributes on an ext2 file system
 */

#ifdef EXT2_SECRM_FL
  if ( (*flags & EXT2_SECRM_FL) != 0  ) { 
    c_attr[0] = 's';
  }
#endif
#ifdef EXT2_UNRM_FL 
  if ( (*flags & EXT2_UNRM_FL) != 0     ) {
    c_attr[1] = 'u';
  }
#endif
#ifdef EXT2_SYNC_FL
  if ( (*flags & EXT2_SYNC_FL) != 0     ) {
    c_attr[2] = 'S';
  }
#endif
#ifdef EXT2_IMMUTABLE_FL
  if ( (*flags & EXT2_IMMUTABLE_FL) != 0   ) {
    c_attr[3] = 'i';
  }
#endif
#ifdef EXT2_APPEND_FL
  if ( (*flags & EXT2_APPEND_FL) != 0    ) {
    c_attr[4] = 'a';
  }
#endif
#ifdef EXT2_NODUMP_FL
  if ( (*flags & EXT2_NODUMP_FL) != 0   ) {
    c_attr[5] = 'd';
  } 
#endif
#ifdef EXT2_NOATIME_FL
  if ( (*flags & EXT2_NOATIME_FL) != 0) {
    c_attr[6] = 'A';
  } 
#endif
#ifdef EXT2_COMPR_FL
  if ( (*flags & EXT2_COMPR_FL) != 0    ) {
    c_attr[7] = 'c';
  } 
#endif
#ifdef EXT2_COMPRBLK_FL
  if ( (*flags & EXT2_COMPRBLK_FL) != 0    ) {
    c_attr[8] = 'B';
  }
#endif
#ifdef EXT2_DIRTY_FL
  if ( (*flags & EXT2_DIRTY_FL) != 0    ) {
    c_attr[9] = 'D';
  }
#endif
#ifdef EXT2_NOCOMPR_FL
  if ( (*flags & EXT2_NOCOMPR_FL) != 0    ) {
    c_attr[10] = 'D';
  }
#endif
#ifdef EXT2_ECOMPR_FL
  if ( (*flags & EXT2_ECOMPR_FL) != 0    ) {
    c_attr[11] = 'D';
  }
#endif

  c_attr[12] = '\0';

  /* fprintf(stderr, "%s  %#x\n", c_attr, f); */

  return (0);
}
#endif

/* determine file type
 */
static 
int sh_unix_getinfo_type (struct stat * buf, 
			  ShFileType * type, 
			  char * c_mode)
{
  if      ( S_ISREG(buf->st_mode)  ) { 
    (*type)   = SH_FILE_REGULAR;
    c_mode[0] = '-';
  }
  else if ( S_ISLNK(buf->st_mode)  ) {
    (*type)   = SH_FILE_SYMLINK;
    c_mode[0] = 'l';
  }
  else if ( S_ISDIR(buf->st_mode)  ) {
    (*type)   = SH_FILE_DIRECTORY;
    c_mode[0] = 'd';
  }
  else if ( S_ISCHR(buf->st_mode)  ) {
    (*type)   = SH_FILE_CDEV;
    c_mode[0] = 'c';
  }
  else if ( S_ISBLK(buf->st_mode)  ) {
    (*type)   = SH_FILE_BDEV;
    c_mode[0] = 'b';
  }
  else if ( S_ISFIFO(buf->st_mode) ) {
    (*type)   = SH_FILE_FIFO;
    c_mode[0] = '|';
  }
  else if ( S_ISSOCK(buf->st_mode) ) {
    (*type)   = SH_FILE_SOCKET;
    c_mode[0] = 's';
  }
  else                              {
    (*type)   = SH_FILE_UNKNOWN;
    c_mode[0] = '?';
  }

  return (0);
}


static 
int  sh_unix_getinfo_mode (struct stat *buf, 
			   unsigned int * mode, 
			   char * c_mode)
{
  (*mode) = buf->st_mode;

  /* make 'ls'-like string */
  
  if ( (buf->st_mode & S_IRUSR) != 0 )  c_mode[1] = 'r'; 
  if ( (buf->st_mode & S_IWUSR) != 0 )  c_mode[2] = 'w'; 
  if ( (buf->st_mode & S_IXUSR) != 0 ) {
    if ((buf->st_mode & S_ISUID) != 0 ) c_mode[3] = 's';
    else                                c_mode[3] = 'x';
  } else {
    if ((buf->st_mode & S_ISUID) != 0 ) c_mode[3] = 'S';
  }

  if ( (buf->st_mode & S_IRGRP) != 0 )  c_mode[4] = 'r'; 
  if ( (buf->st_mode & S_IWGRP) != 0 )  c_mode[5] = 'w'; 
  if ( (buf->st_mode & S_IXGRP) != 0 )  {
    if ((buf->st_mode & S_ISGID) != 0 ) c_mode[6] = 's';
    else                                c_mode[6] = 'x';
  } else {
    if ((buf->st_mode & S_ISGID) != 0 ) c_mode[6] = 'S';
  } 

  if ( (buf->st_mode & S_IROTH) != 0 )  c_mode[7] = 'r'; 
  if ( (buf->st_mode & S_IWOTH) != 0 )  c_mode[8] = 'w';
#ifdef S_ISVTX  /* not POSIX */
  if ( (buf->st_mode & S_IXOTH) != 0 )  {
    if ((buf->st_mode & S_ISVTX) != 0 ) c_mode[9] = 't';
    else                                c_mode[9] = 'x';
  } else {
    if ((buf->st_mode & S_ISVTX) != 0 ) c_mode[9] = 'T';
  }
#else
  if ( (buf->st_mode & S_IXOTH) != 0 )  c_mode[9] = 'x';
#endif 

  return (0);
}

/* #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) */
#endif

void sh_unix_closeall (int fd)
{
#ifdef OPEN_MAX
  int fdlimit = OPEN_MAX;
#else
  int fdlimit = sysconf(_SC_OPEN_MAX);
#endif

  /* can't happen - so fix it :-(
   */
  if (fdlimit < 0)
    fdlimit = 8;  /* POSIX lower limit */

  if (fdlimit > 65536)
    fdlimit = 65536;


  while (fd < fdlimit)
    close(fd++);
  return;
}

static void sh_unix_setlimits()
{
  struct rlimit limits;

  limits.rlim_cur = RLIM_INFINITY;
  limits.rlim_max = RLIM_INFINITY;

#ifdef RLIMIT_CPU
  setrlimit (RLIMIT_CPU,     &limits);
#endif
#ifdef RLIMIT_FSIZE
  setrlimit (RLIMIT_FSIZE,   &limits);
#endif
#ifdef RLIMIT_DATA
  setrlimit (RLIMIT_DATA,    &limits);
#endif
#ifdef RLIMIT_STACK
  setrlimit (RLIMIT_STACK,   &limits);
#endif
#ifdef RLIMIT_RSS
  setrlimit (RLIMIT_RSS,     &limits);
#endif
#ifdef RLIMIT_NPROC
  setrlimit (RLIMIT_NPROC,   &limits);
#endif
#if defined(RLIMIT_NOFILE)
  setrlimit (RLIMIT_NOFILE,  &limits);
#elif defined(RLIMIT_OFILE)
  setrlimit (RLIMIT_OFILE,   &limits);
#endif
#ifdef RLIMIT_MEMLOCK
  setrlimit (RLIMIT_MEMLOCK, &limits);
#endif

  /* no core dumps
   */
  limits.rlim_cur = 0;
  limits.rlim_max = 0;
#ifdef RLIMIT_CORE
  setrlimit (RLIMIT_CORE,    &limits);
#endif

  return;
}

/* delete all environment variables
 */
static void sh_unix_zeroenv()
{
  char * c;
  char ** env = environ;

  while (*env != NULL) {
    c = strchr ((*env), '=');
    ++c;
    ++(env);
    while ((*c) != '\0') {
      (*c) = '\0';
      ++c;
    }
  }
  return;
}


static void  sh_unix_resettimer()
{
  struct itimerval this_timer;

  this_timer.it_value.tv_sec  = 0;
  this_timer.it_value.tv_usec = 0;

  setitimer(ITIMER_REAL,    &this_timer, NULL);
  setitimer(ITIMER_VIRTUAL, &this_timer, NULL);
  setitimer(ITIMER_PROF,    &this_timer, NULL);

  return;
}

static void  sh_unix_resetsignals()
{
  int  sig_num;
#ifdef NSIG
  int  max_sig = NSIG; 
#else
  int  max_sig = 255;
#endif
  int  test;
  struct sigaction act, oldact;
  int  status;

  sigset_t set_proc;

  /* 
   * Reset the current signal mask (inherited from parent process).
   */

  sigfillset(&set_proc);
  EINTR_RETRY(sigprocmask(SIG_UNBLOCK, &set_proc, NULL));

  /* 
   * Reset signal handling.
   */
  
  act.sa_handler = SIG_DFL;         /* signal action           */
  sigemptyset( &act.sa_mask );      /* set an empty mask       */
  act.sa_flags = 0;                 /* init sa_flags           */

  for (sig_num = 1; sig_num <= max_sig; ++sig_num) 
    {
      test = sigaction(sig_num,  &act, &oldact);
      if ((test == -1) && (errno != EINVAL)) 
	{
	  status = errno;
	  sh_error_handle (SH_ERR_WARN, 
			   _("Failed reset of signal %d: %s"),
			   sig_num,
			   sh_error_message (status));
	}
    }

  return;
}

/* Get the local hostname (FQDN)
 */
#include <netdb.h>
#include <sys/socket.h> 

/* Required for BSD
 */
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#include <arpa/inet.h>
#ifdef HAVE_UNAME
#include <sys/utsname.h>
void sh_unix_localhost()
{
  struct utsname   buf;
  struct hostent * he1;
  int              i;
  int              ddot = 0;
  int              len;


  (void) uname (&buf);

  sl_strlcpy (sh.sh_sysname, buf.sysname, SH_MINIBUF+1);
  sl_strlcpy (sh.sh_release, buf.release, SH_MINIBUF+1);
  sl_strlcpy (sh.sh_machine, buf.machine, SH_MINIBUF+1);

  he1 = gethostbyname(buf.nodename);

  if (he1 == NULL)
    sl_strlcpy (sh.localhost, buf.nodename, SH_MINIBUF+1);
  else
    sl_strlcpy (sh.localhost, he1->h_name, SH_MINIBUF+1);
  

  /* check whether it looks like a FQDN
   */
  len = sl_strlen(sh.localhost);
  for (i = 0; i < len; ++i) 
    if (sh.localhost[i] == '.') ++ddot; 

  if (ddot == 0 && he1 != NULL) 
    sl_strlcpy (sh.localhost, 
	     inet_ntoa (*(struct in_addr *) he1->h_addr), 
	     SH_MINIBUF+1); 

  return;
}
#else
void sh_unix_localhost()
{
  struct hostent * he1;

  (void) gethostname (sh.localhost, SH_MINIBUF);
  sh.localhost[SH_MINIBUF] = '\0';
  he1 = gethostbyname(sh.localhost);

  if (he1 != NULL)
    sl_strlcpy (sh.localhost, he1->h_name, SH_MINIBUF+1);
  else
    {
      sl_strlcpy (sh.localhost, _("localhost"), SH_MINIBUF+1);
      return;
    }

  /* check whether it looks like a FQDN
   */
  len = sl_strlen(sh.localhost);
  for (i = 0; i < len; ++i) 
    if (sh.localhost[i] == '.') ++ddot; 
  if (ddot == 0) 
    sl_strlcpy (sh.localhost, 
	     inet_ntoa (*(struct in_addr *) he1->h_addr), 
	     SH_MINIBUF+1); 

  return;
}
#endif


void sh_unix_mlock()
{
#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
  if (skey->mlock_failed == GOOD)
    {
      sl_set_suid ();
      if ( (-1) == mlock( (char *) skey, sizeof (sh_key_t)) ) 
	{
	  skey->mlock_failed = BAD;
	}
      sl_unset_suid ();
    }
#else
  if (skey->mlock_failed == GOOD)
    {
      skey->mlock_failed = BAD;
    }
#endif
}
  
/* daemon mode 
 */
int sh_unix_setdeamon(char * dummy)
{
  static int block = 0;

  if (block != 0)
    return 0;

  if (dummy == NULL)
    sh.modus = 1;
  else if (0 == sl_strncmp(dummy, _("no"), 2))
    sh.modus = 0;
  else
    sh.modus = 1;

  if (sh.opts == 1)  
    block = 1;
	   
  return 0;
}

int sh_unix_init(int goDaemon)
{
  int status;

  /* fork twice, exit the parent process
   */
  if (goDaemon == 1) {
    
    switch (fork()) {
    case 0:  break;      /* child process continues */
    case -1: return -1;  /* error                   */
    default: _exit(0);   /* parent preocess exits   */
    }

    /* Child processes do not inherit page locks across a fork.
     * Error in next fork would return in this (?) thread of execution.
     */
    sh_unix_mlock();
    setsid();            /* should not fail         */

    switch (fork()) {
    case 0:  break;      /* child process continues */
    case -1: return -1;  /* error                   */
    default: _exit(0);   /* parent preocess exits   */
    }

    /* Child processes do not inherit page locks across a fork.
     */
    sh_unix_mlock();

  } else {
    setsid();            /* should not fail         */
  } 

  /* set working directory   
   */
  status = EINTR_RETRY(chdir("/"));
  if ( (-1) == status ) 
    {
      sh_error_handle (SH_ERR_WARN,
		       _("Failed chdir(): %s"),
		       sh_error_message (status));
    }

  /* reset timers 
   */
  sh_unix_resettimer();

  /* signal handlers 
   */
  sh_unix_resetsignals();
  sh_unix_siginstall  ();

  /* set resource limits to maximum, and
   * core dump size to zero 
   */
  sh_unix_setlimits();

  /* set file creation mask 
   */
  (void) umask (0); /* should not fail */

  /* zero out the environment (like PATH='\0')  
   */
  sh_unix_zeroenv();

  /* close all file descriptors, and 
   * open first three streams to /dev/null 
   */
  if (goDaemon == 1)
    {
      sh_unix_closeall (0);
      open(_("/dev/null"), O_RDWR);
      dup(0); 
      dup(0);
    }

  return (0);
}

/* figure out the time offset of the current timeset
 * in a portable way
 */
char * t_zone(const time_t * xx)
{
  struct tm   aa;
  struct tm   bb;
  struct tm * cc;
  int  sign =  0;
  int  diff =  0;
  int  hh, mm;
  static char tz[64];

  cc = gmtime (xx);
  memcpy (&aa, cc, sizeof(struct tm));
  cc = localtime (xx);
  memcpy (&bb, cc, sizeof(struct tm));

  /* check for datum wrap-around
   */
  if      (aa.tm_year < bb.tm_year)
    sign = (-1);
  else if (aa.tm_mon  < bb.tm_mon)
    sign = (-1);
  else if (aa.tm_mday < bb.tm_mday)
    sign = (-1);
  else if (bb.tm_year < aa.tm_year)
    sign = ( 1);
  else if (bb.tm_mon  < aa.tm_mon)
    sign = ( 1);
  else if (bb.tm_mday < aa.tm_mday)
    sign = ( 1);

  diff = aa.tm_hour * 60 + aa.tm_min;
  diff = (bb.tm_hour * 60 + bb.tm_min) - diff;
  diff = diff - (sign * 24 * 60);   /* datum wrap-around correction */
  hh = diff / 60;
  mm = diff - (hh * 60);
  sprintf (tz, "%+03d%02d", hh, mm);
  return (tz);
}

unsigned long sh_unix_longtime ()
{
  time_t        time_now;

  (void) time (&time_now);
  return (unsigned long)time_now;
} 

#ifdef HAVE_GETTIMEOFDAY
unsigned long sh_unix_notime ()
{
  struct timeval  tv;
  struct timezone tz;

  tz.tz_minuteswest = 0;
  tz.tz_dsttime     = 0;

  gettimeofday (&tv, &tz);

  return ((unsigned long)(tv.tv_sec + tv.tv_usec * 10835 + getpid() + getppid()));
  
}
#endif

int sh_unix_settimeserver (char * address)
{
  unsigned long  numcop;

  if (address != NULL) 
    {
      numcop = sl_strlcpy (sh.timeserver, address, 256+1);
      if (SL_ISERROR(numcop)) 
	{
	  fprintf(stderr, 
		  _("ERROR : Time server address too long (max=256): %s"), 
		  address);
	  return (-1);
	}
      return (0);
    }
  return (-1);
}


#ifdef HAVE_NTIME
#define UNIXEPOCH 2208988800UL  /* difference between Unix time and net time 
                                 * The UNIX EPOCH starts in 1970.
                                 */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>
#endif

/* Timeserver service.               */
/* define is missing on HP-UX 10.20  */
#ifndef IPPORT_TIMESERVER 
#define IPPORT_TIMESERVER 37 
#endif


char * sh_unix_time (unsigned long thetime)
{

  int           status;

  time_t        time_now;
  struct tm   * time_ptr;
  static char   AsciiTime[81];                       /* local time   */
  static char   RetTime[81];                         /* local time   */
  static char   deftime[] = N_("[0000-00-00T00:00:00]"); /* default time */


#ifdef HAVE_NTIME
  int    fd;                    /* network file descriptor                  */
  u_char net_time[4];           /* remote time in network format            */
  static int failerr = 0;       /* no net time                              */
  int    fail = 0;              /* no net time                              */
  int    errflag;
  char   errmsg[256];
#endif
  

#ifdef HAVE_NTIME
  if (thetime == 0) 
    {
      if (sh.timeserver[0] == '\0') 
	{
	  fail = 1;
	  (void) time (&time_now);
	} 
      else /* have a timeserver address */
	{ 
	  fd = connect_port (sh.timeserver, IPPORT_TIMESERVER, 
			     errmsg, sizeof(errmsg));
	  if (fd >= 0)
	    {
	      if (4 != read_port (fd, (char *) net_time, 4, &errflag, 2))
		{
		  fail = 1;
		  sl_strlcpy(errmsg, _("read_port(): %s: failed"), 
			     sizeof(errmsg));
		}
	      close(fd);
	    }
	  else
	    fail = 1;
	  
	  if (fail == 0) 
	    { 
	      time_now = ntohl(* (long *) net_time) - UNIXEPOCH;
	      if (failerr == 1) {
		failerr = 0;
		sh_error_handle (SH_ERR_ERR,
				 _("Connection to timeserver %s "\
				   "re-established."),
				 sh.timeserver);
	      } 
	    }
	  else
	    {
	      (void) time (&time_now);
	      if (failerr == 0)
		{
		  failerr = 1;
		  sh_error_handle (SH_ERR_ERR, 
				   errmsg,
				   sh.timeserver);
		}
	    }
	}
    }
  else 
    {
      time_now = thetime;
    }

  /* #ifdef HAVE_NTIME */
#else

  if (thetime == 0) 
    {
      (void) time (&time_now);
    } 
  else 
    {
      time_now = thetime;
    }

  /* #ifdef HAVE_NTIME */
#endif

  if (time_now == (-1) ) 
    return _(deftime);
  else
    time_ptr   = localtime (&time_now);

  if (time_ptr != NULL) 
    {
      status = strftime (AsciiTime, 80, _("[%Y-%m-%dT%H:%M:%S%%s]"), 
			 time_ptr);

      sl_snprintf(RetTime, 80, AsciiTime, t_zone(&time_now));

      if ( (status == 0) || (status == 80) ) 
	return _(deftime);
      else
	return &RetTime[0];
    }

  /* last resort
   */
  return _(deftime);
}



char *  sh_unix_getUIDdir (int level, uid_t uid)
{
  struct passwd * tempres;

  tempres = getpwuid(uid);

  if (tempres == NULL) {
    sh_error_handle (level, 
		     _("Password file entry is NULL for UID %ld"),
		     (long) uid);
    return NULL;
  }

  if (tempres->pw_dir != NULL) {
      return tempres->pw_dir;
  } else {
    sh_error_handle (level, 
		     _("Password field pw_dir is NULL for UID %ld"),
		     (long) uid);
    return NULL;
  }
}

char *  sh_unix_getUIDname (int level, uid_t uid)
{
  struct passwd * tempres;

  tempres = getpwuid(uid);

  if (tempres == NULL) {
    sh_error_handle (level, 
		     _("Password file entry is NULL for UID %ld"),
		     (long) uid);
    return NULL;
  }

  if (tempres->pw_name != NULL) {
      return tempres->pw_name;
  } else {
    sh_error_handle (level, 
		     _("Password field pw_user is NULL for UID %ld"),
		     (long) uid);
    return NULL;
  }
}

char *  sh_unix_getGIDname (int level, gid_t gid)
{
  struct group * tempres;

  tempres = getgrgid(gid);

  if (tempres == NULL) {
    sh_error_handle (level,
		     _("Group file entry is NULL for GID %ld"),
		     (long) gid);
    return NULL;
  }

  if (tempres->gr_name != NULL) {
      return tempres->gr_name;
  } else {
    sh_error_handle (level, 
		     _("Group field pw_user is NULL for GID %ld"),
		     (long) gid);
    return NULL;
  }
}

int sh_unix_getUser ()
{
  struct passwd * tempres;
  char          * p;
  uid_t  seuid, sruid;

  (void) sl_get_euid(&seuid);

  p = sh_unix_getUIDdir (SH_ERR_ERR, seuid);

  if (p == NULL)
    return (-1);
  else
    {
      if (sl_strlen(p) > SH_BUFSIZE) {
	sh_error_handle (SH_ERR_ERR, 
			 _("EUID home directory path is too long (%s)"), 
			 p);
	return (-1);
      } else {
	sl_strlcpy ( sh.ehome, p, SH_BUFSIZE+1);
      }
    }

  (void) sl_get_ruid(&sruid);

  p = sh_unix_getUIDname (SH_ERR_ERR, sruid);
  if (p == NULL)
    return (-1);
  else
    {
      if (sl_strlen(p) > SH_MINIBUF) {
	sh_error_handle (SH_ERR_ERR, 
			 _("Username (%s) is too long"), 
			 p );
	return (-1);
      } else {
	sl_strlcpy ( sh.user, p, SH_MINIBUF+1);
      }
    }

  p = sh_unix_getUIDdir (SH_ERR_ERR, sruid);

  if (p == NULL)
    return (-1);
  else
    {
      if (sl_strlen(p) > SH_BUFSIZE) {
	sh_error_handle (SH_ERR_ERR, 
			 _("Home directory path is too long (%s)"), 
			 p);
	return (-1);
      } else {
	sl_strlcpy ( sh.home, p, SH_BUFSIZE+1);
      }
    }

  tempres = getpwuid(sruid);

  if (tempres->pw_gecos != NULL) {
    if (sl_strlen(tempres->pw_gecos) > SH_MINIBUF) {
      sh_error_handle (SH_ERR_ERR,
		       _("Gecos field (%s) is too long"), 
		       tempres->pw_gecos );
      return (-1);
    } else {
      sl_strlcpy ( sh.real_user, tempres->pw_gecos, SH_MINIBUF+1);
    }
  } else {
    sh_error_handle (SH_ERR_WARN,  
		     _("Password field pw_gecos is NULL") );
    sh.real_user[0] = '\0';
  }

  return (0);

  /* notreached */
}



int sh_unix_getline (SL_TICKET fd, char * line, int sizeofline)
{
  register int  count;
  register int  n = 0;
  char          c;

  if (sizeofline < 2) {
    line[n] = '\0';
    return 0;
  }

  while (n < sizeofline) {

    count = sl_read (fd, &c, 1);

    /* end of file
     */
    if (count < 1) {
      line[n] = '\0';
      n = -1;
      break;
    } 

    if (c != '\0' && c != '\n') {
      line[n] = c;
      ++n;
    } else if (c == '\n') {
      if (n > 0) {
	line[n] = '\0';
	break;
      } else {
	line[n] = '\n'; /* get newline only if only char on line */
	++n;
	line[n] = '\0';
	break;
      }
    } else {
      line[n] = '\0';
      break;
    }

  }

  line[sizeofline] = '\0';  /* make sure line is terminated */
  return n;
}


#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 

/* obtain file info
 */
int sh_unix_getinfo (int level, 
		     char * filename, file_type * theFile, char * fileHash)
{
  struct stat   buf;
  struct stat   lbuf;
  struct stat   fbuf;
  int           stat_return;
  register int  i;
  ShFileType    type;
  unsigned int  mode;
  char        * name;
  char        * tmp;
  char        * tmp2;
  char        * linknamebuf;
  int           linksize;
  
  int get_the_fd (SL_TICKET ticket);

  SL_TICKET     rval_open;
  int           fd;
  int           fstat_return;

  /* stat the file
   */
  sl_set_suid();
  stat_return = EINTR_RETRY(lstat (theFile->fullpath, &buf));
  sl_unset_suid();

  if (S_ISREG(buf.st_mode))
    rval_open = sl_open_read (theFile->fullpath, SL_YESPRIV);
  else
    rval_open = -1;

  if (SL_ISERROR(rval_open))
    fd = -1;
  else
    fd = get_the_fd(rval_open);
      
  if (fd >= 0) 
    {
      sl_set_suid();
      fstat_return = EINTR_RETRY(fstat (fd, &fbuf));
      sl_unset_suid();
    }
  else
    fstat_return = -1;

  /* ---  case 1: lstat failed  --- 
   */
  if (stat_return != 0) 
    {
      stat_return = errno;
      if (!SL_ISERROR(rval_open))
	  sl_close(rval_open);
      if (sh.checkSum == SH_CHECK_INIT || 
	  0 == sh_hash_have_it (theFile->fullpath))
	{
	  tmp2 = sh_util_safe_name (theFile->fullpath);
	  sh_error_handle (level, 
			   _("No file status: %s: lstat: %s"),
			   tmp2,
			   sh_error_message (stat_return)
			   );
	  SH_FREE(tmp2);
	}
      return (-1);
    }

  /* ---  case 2: not a regular file  --- 
   */
  else if (! S_ISREG(buf.st_mode) )
    {
      if (fileHash != NULL)
	sl_strlcpy(fileHash, 
		   _("000000000000000000000000000000000000000000000000"), 
		   KEY_LEN + 1);
    }
  
  
  /* ---  case 3a: a regular file, fstat ok --- 
   */
  else if (fstat_return == 0 && 
	   buf.st_mode == fbuf.st_mode &&
	   buf.st_ino  == fbuf.st_ino  &&
	   buf.st_uid  == fbuf.st_uid  &&
	   buf.st_gid  == fbuf.st_gid  &&
	   buf.st_dev  == fbuf.st_dev )
    {
      if (fileHash != NULL)
	{
	  tiger_fd = rval_open;
	  sl_strlcpy(fileHash, 
		     sh_tiger_hash (theFile->fullpath, TIGER_FD, 0), 
		     KEY_LEN + 1);
	}
    }

  /* ---  case 3b: a regular file, fstat ok, but different --- 
   */
  else if (fstat_return == 0 && S_ISREG(fbuf.st_mode))
    {
      if (fileHash != NULL)
	{
	  tiger_fd = rval_open;
	  sl_strlcpy(fileHash, 
		     sh_tiger_hash (theFile->fullpath, TIGER_FD, 0), 
		     KEY_LEN + 1);
	}
      memcpy (&buf, &fbuf, sizeof( struct stat ));
    }

  /* ---  case 4: a regular file, fstat failed --- 
   */

  else    /* fstat_return != 0 or !S_ISREG(fbuf->st_mode) */
    {
      fstat_return = errno;
      if (fileHash != NULL)
	sl_strlcpy(fileHash, 
		   _("000000000000000000000000000000000000000000000000"), 
		   KEY_LEN + 1);

      tmp2 = sh_util_safe_name (theFile->fullpath);
      if (fstat_return != 0)
	  sh_error_handle (level, 
			   _("No file status: %s: fstat: %s"),
			   tmp2,
			   sh_error_message (fstat_return)
			   );
      else
	  sh_error_handle (level, 
			   _("Not a regular file: %s"),
			   tmp2
			   );
      SH_FREE(tmp2);
    }	  
      
  if (!SL_ISERROR(rval_open))
    sl_close(rval_open);




#if defined(__linux__)
  for (i = 0; i < 12; ++i) theFile->c_attributes[i] = '-';
  theFile->c_attributes[12] = '\0';
  theFile->attributes      =    0;
  sh_unix_getinfo_attr(theFile->fullpath, 
		       &theFile->attributes, theFile->c_attributes);
#endif

  for (i = 0; i < 10; ++i) theFile->c_mode[i] = '-';
  theFile->c_mode[10] = '\0';
  for (i = 0; i < 11; ++i) theFile->link_c_mode[i] = '-';
  theFile->link_c_mode[10] = '\0';


  /* determine file type
   */
  sh_unix_getinfo_type (&buf, &type, theFile->c_mode);
  theFile->type = type;

  /* determine permissions
   */
  sh_unix_getinfo_mode (&buf, &mode, theFile->c_mode);
  theFile->mode = mode;

  /* trivia
   */
  theFile->dev       = (unsigned long) buf.st_dev;
  theFile->ino       = (unsigned long) buf.st_ino;
  theFile->mode      = (unsigned long) buf.st_mode;
  theFile->hardlinks = (unsigned long) buf.st_nlink;
  theFile->owner     = (long)          buf.st_uid;  
  theFile->group     = (long)          buf.st_gid;  
  theFile->rdev      = (unsigned long) buf.st_rdev;
  theFile->size      = (unsigned long) buf.st_size;
  theFile->blksize   = (unsigned long) buf.st_blksize;
  theFile->blocks    = (unsigned long) buf.st_blocks;
  theFile->atime     = (unsigned long) buf.st_atime;
  theFile->mtime     = (unsigned long) buf.st_mtime;
  theFile->ctime     = (unsigned long) buf.st_ctime;


  /* owner and group
   */

  if ( (name = sh_unix_getGIDname(level, buf.st_gid)) != NULL) {
    sl_strlcpy (theFile->c_group, name, GROUP_MAX+1);
  } else {
    tmp2 = sh_util_safe_name (theFile->fullpath);
    sh_error_handle (level, 
		     _("No group name for file: %s"),
		     tmp2);
    SH_FREE(tmp2);
    theFile->c_group[0] = '\0';
  }

  
  if ( (name = sh_unix_getUIDname(level, buf.st_uid)) != NULL) {
    sl_strlcpy (theFile->c_owner, name, OWNER_MAX+1);
  } else {
    tmp2 = sh_util_safe_name (theFile->fullpath);
    sh_error_handle (level,
		     _("No owner name for file: %s"),
		     tmp2);
    SH_FREE(tmp2);
    theFile->c_owner[0] = '\0';
  }

  /* output the file
   */
  tmp2 = sh_util_safe_name (filename);
  sh_error_handle (SH_ERR_ALL,
		   _("%10s %2d %8s %8s %14ld %21s %s"), 
		   theFile->c_mode,
		   theFile->hardlinks,
		   theFile->c_owner,
		   theFile->c_group,
		   theFile->size,
		   sh_unix_time (theFile->mtime),
		   tmp2);
  SH_FREE(tmp2);

  /* check for links
   */
  if (theFile->c_mode[0] == 'l') {

    linknamebuf = SH_ALLOC(PATH_MAX + 2);

    linksize    = readlink (theFile->fullpath, linknamebuf, PATH_MAX);
    if (linksize < PATH_MAX) linknamebuf[linksize] = '\0';
    else linknamebuf[PATH_MAX] = '\0';

    if (linksize < 0) {
      linksize = errno;
      tmp2 = sh_util_safe_name (theFile->fullpath);
      sh_error_handle (level, 
		       _("readlink: file %s: %s"),
		       tmp2,
		       sh_error_message (linksize));
      SH_FREE(tmp2);
      SH_FREE(linknamebuf);
      return (-1);
    }
    
    if (linknamebuf[0] == '/') {
      sl_strlcpy (theFile->linkpath, linknamebuf, PATH_MAX+1);
    } else {
      tmp = sh_util_basename(theFile->fullpath);
      sl_strlcpy (theFile->linkpath, 
		  tmp, 
		  PATH_MAX+1);
      SH_FREE(tmp);
      sl_strlcat (theFile->linkpath, 
	       "/", PATH_MAX+1);
      sl_strlcat (theFile->linkpath, 
	       linknamebuf, 
	       PATH_MAX+1);
    }

    /* stat the link
     */
    sl_set_suid();
    stat_return = EINTR_RETRY(lstat (theFile->linkpath, &lbuf)); 
    sl_unset_suid();

    /* check for error
     */
    if (stat_return != 0) { 
      stat_return = errno;
      tmp  = sh_util_safe_name (theFile->fullpath);
      tmp2 = sh_util_safe_name (theFile->linkpath);
      if (errno != ENOENT) 
	sh_error_handle (level,
			 _("Could not check linked file %s: %s"),
			 tmp2,
			 sh_error_message (stat_return)
			 );
      else /* a dangling link -- everybody seems to have plenty of them */
	sh_error_handle (SH_ERR_INFO,
			 _("Dangling link %s -> %s: %s"),
			 tmp,
			 tmp2,
			 sh_error_message (stat_return)
			 );
      theFile->linkisok = BAD;
      SH_FREE(tmp);
      SH_FREE(tmp2);
      SH_FREE(linknamebuf);
      return (-1);
    }

    theFile->linkisok = GOOD;
      

    /* determine file type
     */
    sh_unix_getinfo_type (&lbuf, &type, theFile->link_c_mode);
    theFile->type = type;
  
    /* determine permissions
     */
    sh_unix_getinfo_mode (&lbuf, &mode, theFile->link_c_mode);
    theFile->linkmode = mode;

    /* output the link
     */
    if (theFile->linkisok == GOOD) {
      tmp2 = sh_util_safe_name (linknamebuf);      
      sh_error_handle (SH_ERR_ALL,
		       _("   >>>  %10s  %s"),
		       theFile->link_c_mode,
		       tmp2);
      SH_FREE(tmp2);
    }

    SH_FREE(linknamebuf);

  }
  
  return (0);
}

/*  #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)  */
#endif


/* check whether file is locked
 */
int sh_unix_testlock (char * filename, int level)
{
  static        struct stat   buf;
  int           status = 0;

  char        * tmp;
  SL_TICKET     fd;
  char          line_in[128];

  /*
  if (sh.firstaccess == 0)
    return 0;
  */

  sl_set_suid();
  status = EINTR_RETRY(lstat (filename, &buf));
  sl_unset_suid();

  if (status < 0)
    return (-1);

  sl_set_suid();
  status = EINTR_RETRY(lstat (DEFAULT_ERRLOCK, &buf));
  sl_unset_suid();

  if (status < 0 && errno == ENOENT)
    {
      if (0 == sh_unix_lock ())
	{
	  sh.firstaccess = 0;
	  return (0);
	}
    }

  /* check on lock
   */
  
  if (status >= 0)
    fd = sl_open_read (DEFAULT_ERRLOCK, SL_YESPRIV);
  else
    fd = -1;

  if (!SL_ISERROR(fd))
    {

      /* read the PID in the lock file
       */
      status = sh_unix_getline (fd, line_in, sizeof(line_in)-1);

      /* convert to numeric
       */
      if (status > 0)
	{
	  errno  = 0;
	  status = strtol(line_in, (char **)NULL, 10);
	  if (errno == ERANGE || status <= 0)
	    status = -1;
	}

      if (status == (int) getpid())
	{
	  sl_close(fd);
	  return 0;
	}

      /* check whether the process exists
       */
      if (status > 0)
	{
	  errno  = 0;
	  status = kill (status, 0);

	  /* does not exist, so remove the stale lock
	   * and create a new one
	   */
	  if (status < 0 && errno == ESRCH)
	    {
	      sh.locked = GOOD;
	      if (0 == sh_unix_unlock() &&
		  0 == sh_unix_lock  ())
		{
		  sh.firstaccess = 0;
		  sl_close(fd);
		  return (0);
		}
	    }
	}
      sl_close(fd);
    }

  tmp = sh_util_safe_name (DEFAULT_ERRLOCK);
  sh_error_handle (level,
		   _("Cannot create lock file %s"),
		   tmp);
  SH_FREE(tmp);
  return (-1);
}

int sh_unix_unlock()
{
  /* struct stat buf; */
  int         error = 0;
  
  /* logfile is not locked to us
   */
  if (sh.locked == BAD) 
    return (-1);

  /* check whether the directory is secure
   */
  if (0 != tf_trust_check (DEFAULT_ERRLOCK, SL_YESPRIV))
    return (-1);

  /* delete the lock file 
   */
  sl_set_suid(); 
  error = EINTR_RETRY(unlink (DEFAULT_ERRLOCK));
  sl_unset_suid(); 
  
  if (error == 0)
    sh.locked = BAD; /* not locked anymore */
  else
    {
      error = errno;
      sh_error_handle (SH_ERR_ERR, 
		       _("Error deleting lock file %s: %s"),
		       DEFAULT_ERRLOCK, sh_error_message(error)
		       );
      return (-1);
    }

  return (0);
}

int sh_unix_lock ()
{
  struct stat buf;
  int status;
  int errnum;
  char myPid[64];
  SL_TICKET  fd;

  sl_set_suid(); 
  status = EINTR_RETRY(lstat (DEFAULT_ERRLOCK, &buf));
  sl_unset_suid(); 
  
  if (0 == status) 
    {
      fprintf(stderr, _("Startup error: lockfile %s already exists\n"), 
	      DEFAULT_ERRLOCK);
      sh.locked       = BAD;
      return (-1);
    }

  sprintf (myPid, "%ld\n", (long) getpid());           /* known to fit  */

  fd = sl_open_write (DEFAULT_ERRLOCK, SL_YESPRIV);

  if (!SL_ISERROR(fd))
    {
      errnum = sl_write (fd, myPid, sl_strlen(myPid));
      if (!SL_ISERROR(errnum))
	errnum = sl_close (fd);
      if (!SL_ISERROR(errnum))
	{
	  sh.locked = GOOD;
	  return (0);
	}
    }

  fprintf(stderr, _("Startup error: error creating lockfile %s\n"), 
	  DEFAULT_ERRLOCK);
  sh.locked       = BAD;
  return (-1);

  /* notreached */
}

/* Test whether file exists, is a character device, and allows read
 * access.
 */
int sh_unix_file_exists(char * file_name)
{
  
  struct stat buf;
  if (EINTR_RETRY(stat(file_name, &buf)) == -1)
    return (-1);
  else if ( S_ISCHR(buf.st_mode) &&  0 != (S_IROTH & buf.st_mode)) 
    return (0);
  else 
    return (-1);
}

static char preq[16];

/* return the path to the configuration/database file
 */
char * file_path(char what, char flag)
{
  static int init = 0;

  if (init == 0)
    {
      sl_strlcpy(preq, _("REQ_FROM_SERVER"), 16);
      ++init;
    }

  switch (what)
    {

    case 'C':
      if (0 == sl_strncmp (sh.conf_path, preq, 15))
	{
#if defined(SH_WITH_SERVER)
	  if (sh.isserver == GOOD && sl_strlen(sh.conf_path) == 15)
	    return NULL;
	  if (sh.isserver == GOOD)
	    return &(sh.conf_path[15]);
#endif
	  if (flag == 'R')
	    return preq;
	  if (flag == 'I')
	    {
	      if (sl_strlen(sh.conf_path) == 15)
		return NULL;
	      else
		return &(sh.conf_path[15]);
	    }
	  return preq;
	}
      else
	return sh.conf_path;
      break;

    case 'D':
      if (0 == sl_strncmp (sh.data_path, preq, 15))
	{
	  if (flag == 'R')
	    return preq;
	  if (flag == 'W' && sl_strlen(sh.data_path) == 15)
	    return NULL;
	  if (flag == 'W')
	    return &(sh.data_path[15]);
	}
      else
	return sh.data_path;
      break;
	
    default:
      return NULL;
      break;
    }

  return NULL; /* notreached */
}

/************************************************/
/************************************************/
/****   Stealth Utilities                    ****/
/************************************************/
/************************************************/
#ifdef SH_STEALTH

void sh_unix_xor_code (char * str, int len)
{
  register int i;

  for (i = 0; i < len; ++i) str[i] ^= (char) XOR_CODE;
  return;
}

#if  !defined(SH_STEALTH_MICRO)

static unsigned long off_data = 0;
static unsigned long max_data = 0;
static int           stealth_init = BAD;

int hideout_hex_block(SL_TICKET fd, unsigned char * str, int len);
unsigned long first_hex_block(SL_TICKET fd, unsigned long * max);

/*
 * get hidden data from a block of hex data
 */
int sh_unix_getline_stealth (SL_TICKET fd, char * str, int len)
{
  int add_off;

  /* initialize
   */
  if (stealth_init == BAD)
    {
      off_data = first_hex_block(fd, &max_data);
      if (off_data == 0)
	{
	  sh_error_handle (SH_ERR_FATAL,
			   _("PANIC -- no data in stealth config file."));
	  exit (EXIT_FAILURE);
	}
      stealth_init = GOOD;
    }
  
  /* seek to proper position
   */
  sl_seek(fd, off_data);
     
  /* read one line
   */
  add_off = hideout_hex_block(fd, str, len);
  /*    fprintf(stdout, "%s", str); */
  off_data += add_off;


  if (0 != sl_strncmp(str, _("[EOF]"), 5))
    return sl_strlen(str);
  else
    return 0;

}

int hideout_hex_block(SL_TICKET fd, unsigned char * str, int len)
{

  register int  i, j, k;
  unsigned char c, e;
  register int  num;
  char          mask[9] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
  unsigned long here   = 0;
  unsigned long retval = 0;

  i = 0;
  while (i < len)
    {
      for (j = 0; j < 8; ++j)
	{

	  /* get a low byte, modify, read back */
	  for (k = 0; k < 2; ++k)
	    {
	      c = ' ';
	      do {
		do {
		  num = sl_read (fd, &c, 1);
		} while (num == 0 && errno == EINTR);
		if (num == 0) return -1;
		++here; 
	      } while (c == '\n' || c == '\t' || c == '\r' || 
		       c == ' ');
	    }
	  

	  /* e is the value of the low byte
	   */
	  e = (unsigned char) sh_util_hexchar( c );
	  if ((e & mask[7]) != 0)  /* bit is set     */
	    str[i] |= mask[j];
	  else                     /* bit is not set */
	    str[i] &= ~mask[j];

	}
      if (str[i] == '\n') break;
      ++i;
    }
  if (i != 0)
    str[i] = '\0';
  else
    str[i+1] = '\0';
  retval += here;
  return retval;
}

/* get offset of first data block
 */
unsigned long first_hex_block(SL_TICKET fd, unsigned long * max)
{
  int           i;
  register int  num = 1;
  char          c;
  int           nothex = 0;
  unsigned long retval = 0;
  int           this_line = 0;
  char          theline[SH_BUFSIZE];

  *max = 0;

  while (1)
    {
      theline[0] = '\0';
      this_line  = 0;
      c          = '\0';
      while (c != '\n' && num > 0)
	{
	  do {
	    num = sl_read (fd, &c, 1);
	  } while (num == 0 && errno == EINTR);
	  if (num > 0) theline[this_line] = c;
	  else         return 0;
	  this_line += num;
	}
      theline[this_line] = '\0';
      
      /* not only 'newline' */ 
      if (this_line > 60)
	{
	  nothex  = 0;
	  i       = 0;
	  while (nothex == 0 && i < (this_line-1))
	    {
	      if (! isxdigit((int)theline[i])) nothex = 1;
	      ++i;
	    }
	  if (nothex == 1) retval += this_line;
	}
      else
	{
	  nothex = 1;
	  retval += this_line;
	}

      if (nothex == 0)
	{
	  *max = 0; 
	  do {
	    do {
	      num = sl_read (fd, theline, SH_BUFSIZE);
	    } while (num == 0 && errno == EINTR);
	    for (i = 0; i < num; ++i)
	      { 
		c = theline[i];
		if (c == '\n' || c == '\t' || c == '\r' || c == ' ') 
		  ;
		else if (!isxdigit((int)c))
		  break;
		else
		  *max += 1;
	      }
	  } while (num > 0);

	  *max /= 16;
	  return retval;
	}

    }
  return 0;
}

 /* if !defined(SH_STEALTH_MICRO) */
#endif 

 /* ifdef SH_STEALTH */
#endif  
